<?php

use Aws\CommandPool;
use Aws\S3\Exception\S3Exception;
use Aws\S3\S3Client;

require_once IOWD_DIR_INCLUDES . "/iowd-optimize.php";
require_once IOWD_DIR_INCLUDES . "/iowd-helper.php";
require_once IOWD_DIR_INCLUDES . "/iowd-filter-images.php";
require_once IOWD_DIR_INCLUDES . "/iowd-image-processing.php";
require_once IOWD_DIR_CLASSES . "/iowd-log.php";


class IOWD_Rest extends WP_REST_Controller
{
    private $version = '1';
    private $route = 'iowd';
    private $region = 'us-west-2';
    private $status = 404;
    private $response = array();

    private $bases = array(
        'images\/download' => array('route'   => 'images/download',
                                    'methods' => WP_REST_Server::CREATABLE,
                                    'args'    => array(
                                        'bucket'      => array(
                                            'required' => true,
                                            'type'     => 'string',
                                        ),
                                        'post_id'     => array(
                                            'required' => true,
                                        ),
                                        'iteration'   => array(
                                            'required' => true,
                                        ),
                                        'images_data' => array(
                                            'required' => true
                                        ),
                                    )),
        '[0-9]+\/info'     => array('route'   => '/(?P<domain_id>\d+)/info/',
                                    'methods' => WP_REST_Server::READABLE,
                                    'args'    => array()),
        '[0-9]+\/optimize' => array('route'   => '/(?P<domain_id>\d+)/optimize/',
                                    'methods' => WP_REST_Server::READABLE,
                                    'args'    => array()),
        '[0-9]+\/settings' => array('route'   => '/(?P<domain_id>\d+)/settings/',
                                    'methods' => WP_REST_Server::READABLE,
                                    'args'    => array()),
        '[0-9]+\/cancel'   => array('route'   => '/(?P<domain_id>\d+)/cancel/',
                                    'methods' => WP_REST_Server::READABLE,
                                    'args'    => array()),
        'settings'         => array('route'   => '/settings/',
                                    'methods' => WP_REST_Server::CREATABLE,
                                    'args'    => array(
                                        'optimization_mode'                 => array(
                                            'required' => true,
                                            'type'     => 'string',
                                        ),
                                        'enable_auto_optimization'          => array(
                                            'required' => true,
                                            'type'     => 'int',
                                        ),
                                        'enable_photo_gallery_optimization' => array(
                                            'required' => true,
                                            'type'     => 'int',
                                        ),
                                        'domain_id'                         => array(
                                            'required' => true,
                                            'type'     => 'int',
                                        ),
                                    )),
        'scan'             => array('route'   => '/scan/',
                                    'methods' => WP_REST_Server::CREATABLE,
                                    'args'    => array(
                                        'limit' => array(
                                            'required' => true,
                                            'type'     => 'int',
                                        )
                                    )),
        'log'              => array('route'   => '/log/',
                                    'methods' => WP_REST_Server::CREATABLE,
                                    'args'    => array(
                                        'limit' => array(
                                            'required' => true,
                                            'type'     => 'int',
                                        )
                                    )),

    );


    public function register_routes()
    {
        $namespace = $this->route . '/v' . $this->version;
        foreach ($this->bases as $base => $route_data) {
            register_rest_route($namespace, $route_data['route'], array(
                array(
                    'methods'  => $route_data['methods'],
                    'callback' => array($this, 'callback'),
                    'args'     => $route_data['args']
                )
            ));
        }
    }

    /**
     * @param $request
     *
     * @return WP_REST_Response
     */
    public function callback($request)
    {
        include_once(TENWEB_INCLUDES_DIR . '/class-api.php');
        include_once(TENWEB_INCLUDES_DIR . '/class-tenweb-services.php');
        if (!get_option(IOWD_PREFIX . "_tables_created")) {
            IOWD_Helper::create_tables();
        }

        $route = $request->get_route();
        $domain_id = $request->get_param('domain_id');

        $endpoint = $this->parse_endpoint($route);
        $unauthorized = false;

        $need_10web_authorization = $endpoint == 'images\/download' || $endpoint == 'scan' || $endpoint == 'log' || $endpoint == '[0-9]+\/cancel' ? false : true;
        $need_iowd_authorization = $endpoint == 'scan' || $endpoint == 'log' ? true : false;

        if ($need_10web_authorization === true) {
            if (!defined('TENWEB_INCLUDES_DIR')) {
                $unauthorized = true;
            } else {
                $token = $request->get_header('tenweb-authorization');
                if (empty($token) || \Tenweb_Manager\Api::get_instance()->check_single_token($token) === false) {
                    $unauthorized = true;
                }
            }

            if ($unauthorized === true) {
                $data_for_response = array(
                    "code"    => "unauthorized",
                    "message" => "manager unauthorized, please login",
                    "data"    => array(
                        "status" => 401
                    )
                );

                return new \WP_REST_Response($data_for_response, 401);
            }
        }

        if ($need_iowd_authorization === true) {
            if (!check_ajax_referer('nonce_iowd', 'nonce_iowd', false)) {
                $data_for_response = array(
                    "code"    => "wrong_nonce",
                    "message" => "Wrong nonce.",
                    "data"    => array(
                        "status" => 401
                    )
                );

                //return new \WP_REST_Response($data_for_response, 404);
            }
        }

        $manager_domain_id = \TenwebServices::get_domain_id();

        if (!empty($domain_id)) {
            if ($manager_domain_id != $domain_id) {
                $data_for_response = array(
                    "code"    => "bad_request",
                    "message" => "Domains not match.",
                    "data"    => array(
                        "status" => 400
                    )
                );

                return new \WP_REST_Response($data_for_response, 400);
            }
        }


        try {
            switch ($endpoint) {
                case 'images\/download':
                    $this->images_download($request);
                    break;
                case '[0-9]+\/info':

                    $this->domain_info($request);
                    break;
                case '[0-9]+\/optimize':
                    $this->optimize($request);
                    break;

                case'[0-9]+\/settings':
                    $this->get_settings($request);
                    break;
                case 'settings':
                    $this->store_settings($request);
                    break;
                case 'scan':
                    $this->scan($request);
                    break;
                case 'log':
                    $this->log($request);
                    break;
                case '[0-9]+\/cancel':
                    $this->cancel($request);
                    break;
                default:
                    $this->status = 404;
                    $this->response = array(
                        "code"    => "rest_no_route",
                        "message" => "No route was found matching the URL and request method.",
                        "data"    => array(
                            "status" => 404
                        )
                    );
                    break;
            }
        } catch (\Exception $e) {
            IOWD_Log::set_error_log("REST:error", $e->getMessage() . " on line " . $e->getLine());
            $this->status = 500;
            $this->response = array(
                "code"    => "error",
                "message" => $e->getMessage(),
                "data"    => array(
                    "status" => 500
                )
            );
        }

        return new \WP_REST_Response($this->response, $this->status);
    }

    /**
     * @param WP_REST_Request $request
     *
     * @throws Exception
     */
    public function images_download(WP_REST_Request $request)
    {
        clearstatcache();
        @ini_set('max_execution_time', 1300);
        $bucket = $request->get_param('bucket');
        $images_data = $request->get_param('images_data');
        $post_id = $request->get_param('post_id');
        $iteration = $request->get_param('iteration');
        $credentials = get_option("iowd_crd_" . $post_id);

        IOWD_Log::set_optimization_log("REST:images_download:params", array(
            "iteration"   => $iteration,
            "post_id"     => $post_id,
            "images_data" => $images_data,
        ));

        $s3args = array(
            'version'     => 'latest',
            'region'      => $this->region,
            'credentials' => $credentials,
        );

        $s3client = new S3Client($s3args);

        try {
            $commands = array();
            foreach ($images_data as $image_data) {
                if ($s3client->doesObjectExist($bucket, $image_data["path"])) {
                    $commands[] = $s3client->getCommand('getObject', array(
                        'Bucket' => $bucket,
                        'Key'    => $image_data["path"],
                        'SaveAs' => $image_data["wp_path"],
                    ));
                }
            }
            if (empty($commands) === false) {
                $pool = new CommandPool($s3client, $commands);
                $promise = $pool->promise();
                $promise->wait();
            }
            $optimizer = new IOWD_Optimize();
            $optimizer->save_data_to_db($images_data, $iteration, $post_id);

            $received_images = get_transient("iowd_received_images_" . $post_id);
            if (!$received_images) {
                $data_count = get_option("iowd_data_count_" . $post_id);
                $received_images = range(0, $data_count - 1);
            }
            unset($received_images[$iteration]);

            if (empty($received_images) === true) {

                delete_transient("iowd_received_images_" . $post_id);
                $post_ids = get_transient("iowd_optimizing_post_ids");
                $post_ids = $post_ids ? $post_ids : array();
                if (($key = array_search($post_id, $post_ids)) !== false) {
                    unset($post_ids[$key]);
                    set_transient("iowd_optimizing_post_ids", $post_ids);
                } else {
                    delete_transient("iowd_optimizing_post_ids");
                }
                delete_option("iowd_crd_" . $post_id);
                delete_option("iowd_folder_" . $post_id);
                delete_option("iowd_image_temp_data_" . $post_id);
                delete_option("iowd_data_count_" . $post_id);
                if (strpos($post_id, "auto") === false) {
                    set_transient("iowd_done_" . $post_id, 1);
                }
            } else {
                set_transient("iowd_received_images_" . $post_id, $received_images);
            }

            $this->status = 200;
            $this->response = array('status' => 'ok', 'body' => 'success');

        } catch (S3Exception $e) {
            IOWD_Log::set_error_log("REST:images_download:S3Exception", $e->getMessage() . " on line " . $e->getLine());
            $this->status = 500;
            $this->response = array('status' => 'error', 'body' => $e->getMessage());
        } catch (\Exception $e) {
            IOWD_Log::set_error_log("REST:images_download:Exception", $e->getMessage() . " on line " . $e->getLine());
            $this->status = 500;
            $this->response = array('status' => 'error', 'body' => $e->getMessage());
        }
    }

    /**
     * @param WP_REST_Request $request
     */
    public function domain_info(WP_REST_Request $request)
    {
        require_once IOWD_DIR_INCLUDES . "/iowd-scan.php";
        // get all images
        $scan = new IOWD_Scan();
        // clear cache
        $scan_total = $scan->update_scan_cache(0);

        $this->status = 200;
        $this->response = array('status' => 'ok', "body" => $scan->scanTotal($scan_total));
    }

    /**
     * @param WP_REST_Request $request
     */
    public function optimize(WP_REST_Request $request)
    {
        try {
            require_once IOWD_DIR_INCLUDES . "/iowd-api.php";
            require_once IOWD_DIR_INCLUDES . "/iowd-util.php";
            require_once IOWD_DIR_INCLUDES . "/iowd-amazon-s3.php";

            IOWD_Log::set_optimization_log("REST:optimize:started", "optimization started");

            $optimize = new IOWD_Optimize();
            $data_count = $optimize->get_temp_data(array(), true);
            $data_count = !empty($data_count["data_count"]) ? $data_count["data_count"] : 0;

            IOWD_Log::set_optimization_log("REST:optimize:iterations", $data_count);
            if($data_count == 0){
                $this->response = array('status' => 'no_images', "body" => "No images");
            } else {

                for ($i = 0; $i < $data_count; $i++) {
                    if ($optimize->aborted === true) {
                        break;
                    }
                    $optimize->optimize($i);
                }
                $this->response = array('status' => 'ok', "body" => "All is ok, wait for job response");
            }
            $this->status = 200;
        } catch (\Exception $e) {
            IOWD_Log::set_error_log("REST:optimize", $e->getMessage() . " on line " . $e->getLine());
            $this->status = 500;
            $this->response = array(
                "code"    => "error",
                "message" => $e->getMessage(),
                "data"    => array(
                    "status" => 500
                )
            );
        }
    }

    /**
     * @param WP_REST_Request $request
     */
    public function get_settings(WP_REST_Request $request)
    {
        $data = IOWD_Helper::collect_settings();
        $this->status = 200;
        $this->response = array('status' => 'ok', "body" => $data);
    }

    /**
     * @param WP_REST_Request $request
     */
    public function store_settings(WP_REST_Request $request)
    {
        require_once IOWD_DIR_INCLUDES . "/iowd-settings.php";
        $limitation = IOWD_Helper::limitation();
        $optimization_mode = $request->get_param('optimization_mode');
        $enable_auto_optimization = !empty($limitation["auto_optimize"]) ? $request->get_param('enable_auto_optimization') : 0;
        $enable_photo_gallery_optimization = $request->get_param('enable_photo_gallery_optimization');

        if ($optimization_mode != "extreme" || ($optimization_mode == "extreme" && !empty($limitation["extreme_mode"]))) {
            // save optimization mode
            IOWD_Settings::save_optimization_mode($optimization_mode);
        }

        // save other options
        $settings = json_decode(get_option(IOWD_PREFIX . "_options"), true);
        $settings['automatically_optimize'] = $enable_auto_optimization;
        $settings['optimize_gallery'] = $enable_photo_gallery_optimization;
        update_option(IOWD_PREFIX . "_options", json_encode($settings));

        $this->status = 200;
        $this->response = array('status' => 'ok', "body" => "success");
    }

    public function scan(WP_REST_Request $request)
    {
        $optimizer = new IOWD_Optimize();
        $limit = !empty($request->get_param('limit')) ? $request->get_param('limit') : 0;
        $attachments_count = $optimizer->get_attachments_count();
        $iterations = ceil($attachments_count / 2000);
        $current_iteration = $limit / 2000 + 1;

        $data = $optimizer->get_attachments(array(), false, array(), null, $limit);
        if ($current_iteration < $iterations) {
            $response = array("status" => "pending");
        } else {
            $data = $optimizer->get_attachments(array(), true, array(), null, $limit);
            $response = array("status" => "done");
        }

        $this->status = 200;
        $this->response = $response;
    }

    public function log(WP_REST_Request $request)
    {
        require_once IOWD_DIR_INCLUDES . "/iowd-scan.php";
        $scan = new IOWD_Scan();
        $limit = !empty($request->get_param('limit')) ? $request->get_param('limit') : 0;

        $media = array();
        $other = array();
        $media_sizes = array();
        $attachments = array();
        for ($i = 0; $i <= $limit; $i += 2000) {
            $db_attachments = get_transient("iowd_temp_scan_data_" . $i);
            $media_attachments = isset($db_attachments["media"]) ? $db_attachments["media"] : array();
            $media_sizes_attachments = isset($db_attachments["media_sizes"]) ? $db_attachments["media_sizes"] : array();
            $other_attachments = isset($db_attachments["other"]) ? $db_attachments["other"] : array();
            $u_attachments = isset($db_attachments["attachments"]) ? $db_attachments["attachments"] : array();

            $media = $media + $media_attachments;
            $other = $other + $other_attachments;
            $media_sizes = $media_sizes + $media_sizes_attachments;
            $attachments = $attachments + $u_attachments;

            delete_transient("iowd_temp_scan_data_" . $i);
        }
        $all_attachments = array(
            "media"       => $media,
            "media_sizes" => $media_sizes,
            "other"       => $other,
            "attachments" => $attachments,
        );


        $data = $scan->scan($all_attachments);
        set_transient("iowd_scan_data_cache", $data);
        $this->status = 200;

        $this->response = $data;

    }

    public function cancel(WP_REST_Request $request)
    {
        $optimizer = new IOWD_Optimize();
        $optimizer->abort(false);
        $this->status = 200;
        $this->response = array("status" => "ok");

    }

    /**
     * @param $route
     *
     * @return int|null|string
     */
    private function parse_endpoint($route)
    {
        $route_url = substr($route, 9);
        foreach ($this->bases as $key => $value) {
            $route_regex = '/' . $key . '/';
            if (preg_match($route_regex, $route_url)) {
                return $key;
            }
        }

        return null;
    }

}